use std::io::{stdout, Write};
use ffmpeg::format::Pixel;

#[allow(dead_code)]
pub enum LinesFormat {
    /// All lines go to the right.
    Right,
    /// The first line goes to the right, the next to the left.
    RightLeft,
    /// The first line goes to the left, the next to the right.
    LeftRight,
    /// All lines go to the left.
    Left,
}

#[derive(Debug)]
pub enum ParseError {
    NotRGB24,
    NotSquareAspectRatio,
    VideoIndexOutOfBounds,
}

impl std::fmt::Display for ParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let s = match self {
            ParseError::NotRGB24 => "Not in RGB24 format!",
            ParseError::NotSquareAspectRatio => "Not in a square (1:1) aspect ratio!",
            ParseError::VideoIndexOutOfBounds => "Selected video index is not valid (Out of Bounds)!",
        };
        write!(f, "{s}")
    }
}

impl std::error::Error for ParseError {} 

/// Must use square images
/// Images must be in RGB24
pub fn print_square(frame: &ffmpeg::frame::Video, lines_format: LinesFormat, video_index: usize) -> Result<(), ParseError> {
    if frame.format() != Pixel::RGB24   { return Err(ParseError::NotRGB24); }
    if frame.width() != frame.height()  { return Err(ParseError::NotSquareAspectRatio); } 
    // 8 is the value hard-coded in AVFrame
    if !(0..=8).contains(&video_index)  { return Err(ParseError::VideoIndexOutOfBounds); } 

    let pixels_wide = frame.width() as usize;

    let mut stdout = stdout();
    let mut lock = stdout.lock();

    // EXPLANATION: (12x12 image)
    // The full buffer will be 1152 bytes.
    //
    // 1152/12=96   Buffer length   / Pixels high       = 96 Bytes per line.
    // 96/12=8      Bytes per line  / Pixels per line   = 8 Bytes per pixel.
    //
    // But RGB24 should only use 3 bytes per pixel. The other 5 we just discard.
    // They might be alpha, or grayscale overlay, I'm not really sure.
    //
    // 8-5=3        Bytes per pixel - Discarded bytes = Used bytes
    // 3*12=36      Used bytes      * Pixels per line = Slice length
    //
    // This is because all the extra data (the 5 extra bytes) are at the end
    // of each line. So we just take the first (Pixel width)*3 bytes.

    let buf = frame.data(video_index); 
    let step = buf.len() / frame.height() as usize;

    // (Assuming square) Step thru buffer.
    for i in 0..pixels_wide {
        let j = i*step;
        // Have to take ownership, by implicilty Cloning the data. (Copying?)
        let mut slice = Vec::from(&buf[j..j+pixels_wide*3]);

        match lines_format {
            LinesFormat::RightLeft =>  if i & 1 == 1 { rev_in_place_with_groupings(&mut slice) },
            LinesFormat::LeftRight  => if i & 1 != 1 { rev_in_place_with_groupings(&mut slice) },
            LinesFormat::Left       => rev_in_place_with_groupings(&mut slice),
            LinesFormat::Right      => {},
        }

        // DEBUG
        // println!("{:?} || {}-{}", slice, j, j+step);
        // NORMAL
        lock.write(&slice).unwrap();
    }
    stdout.flush().unwrap();
    Ok(())
}

fn rev_in_place_with_groupings(raw: &mut Vec<u8>) {
    for i in 0..raw.len() {
        match i % 3 {
            // Think of these as the indexes of each
            // RGB tuple. [0, 1, 2] [R, G, B].
            // 0 and 2 have to be switched, while 1 can stay.
            2 => {
                let zero = raw[i-2];
                let two  = raw[i];
                raw[i-2] = two;
                raw[i] = zero;
            },
            _ => { /* 0 and 1*/ },
        } 
    }
}

